local l3d  = require "love3d"
local cpml = require "cpml"
local iqm  = require "iqm"

local models, shader, vignette, noise, start = {}, false, false, false, 0

function love.load()
	-- Monkey-patch the love API - you can use the lib without this by omitting
	-- "true", but it makes things much easier to use.
	l3d.import(true)

	-- Load all the textures and set filtering. If you don't set the filtering
	-- everything looks awful at a distance. Use 8x or higher aniso.
	local flags = {
		srgb = select(3, love.window.getMode()).srgb,
		mipmaps = true
	}

	-- Usually you'd define this stuff in some external file.
	local data = {
		{
			model = "aesa/aesa.iqm",
			use_fresnel = true,
			textures = {
				body = "aesa/skin.png",
				dress = "aesa/dress.png"
			}
		}, {
			model = "aesa/ground.iqm",
			textures = {
				ground = "aesa/ground.png"
			}
		}
	}

	for _, params in ipairs(data) do
		local t = love.timer.getTime()
		local model = iqm.load(params.model)
		print(string.format("Loaded IQM in %0.6fs", love.timer.getTime() - t))

		model.fresnel = params.use_fresnel and 1 or 0
		model.textures = {}
		for material, path in pairs(params.textures) do
			model.textures[material] = love.graphics.newImage(path, flags)
			model.textures[material]:setFilter("linear", "linear", 16)
		end

		table.insert(models, model)
	end

	-- Basic flat shader with support for skinning.
	shader   = love.graphics.newShader("shader.glsl")

	-- Things to pretty the demo up just a little
	-- We mix a little noise in to break up banding. It's a lame way to dither.
	vignette = love.graphics.newShader("vignette.glsl")
	noise    = love.graphics.newImage("textures/noise.png", { srgb = flags.srgb })
	noise:setWrap("repeat", "repeat")
	start    = love.timer.getTime()

	-- Gamma-correct the background color, if needed.
	local cc = flags.srgb and love.math.gammaToLinear or function(t) return t end
	love.graphics.setBackgroundColor(cc(237, 241, 246))
end

function love.draw()
	love.graphics.push("all")

	-- Love doesn't use the depth buffer, so we need to set it up & clear it.
	-- We need depth testing so we don't have to depth-sort opaque objects.
	love.graphics.clearDepth()
	love.graphics.setDepthTest("less")


	-- Now we transform & update the shader params...
	-- World transform. love.graphics.{translate, rotate, scale} are hooked
	local speed = 0.25
	love.graphics.setShader(shader)
	love.graphics.rotate((love.timer.getTime() - start) * speed + math.pi / 2, { 0, 0, -1 })
	shader:send("u_model", love.graphics.getMatrix():to_vec4s())

	-- Camera transform
	local v = cpml.mat4():look_at(cpml.vec3(0, 5, 1.5), cpml.vec3(0, 0, 1), cpml.vec3(0, 0, 1))
	shader:send("u_view", v:to_vec4s())

	-- Projection (camera perspective)
	local w, h = love.graphics.getDimensions()
	local p = cpml.mat4():perspective(25, w/h, 0.01, 1000)
	shader:send("u_projection", p:to_vec4s())

	-- You can draw the whole model as one mesh (just don't set draw range), but
	-- if you draw this way you can assign a different shader/textures per-mesh.
	-- (naturally, at the expense of more draw calls - do what's best for you)
	for _, model in ipairs(models) do
		shader:sendInt("use_fresnel", model.fresnel)
		for _, buffer in ipairs(model) do
			local texture = model.textures[buffer.material]
			model.mesh:setTexture(texture) -- nil is OK
			model.mesh:setDrawRange(buffer.first, buffer.last)
			love.graphics.draw(model.mesh)
		end
	end

	-- Make sure to reset depth testing or drawing anything in 2D-space breaks.
	-- You'd need to reset blending if you disabled that (for, say, FXAA), too.
	love.graphics.setDepthTest()
	love.graphics.setShader()
	love.graphics.pop()

	love.graphics.setShader(vignette)
	love.graphics.setBlendMode("multiply")
	love.graphics.setColor(255, 255, 255, 255)
	love.graphics.draw(noise, 0, 0, 0, w / noise:getWidth(), h / noise:getHeight())
	love.graphics.setBlendMode("alpha")
	love.graphics.setShader()

	love.graphics.setColor(0, 0, 0, 255)
	love.graphics.print(string.format("FPS: %0.2f (%0.3fms)", love.timer.getFPS(), love.timer.getAverageDelta()), 5, 5)
	love.graphics.print("Press ESCAPE to quit.", 5, 25)
end

function love.keypressed(k, s, r)
	if k == "escape" then love.event.quit() end
end
